/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#include "MixCastSdkActor.h"
#include "MixCast.h"
#include "Data/SdkData.h"
#include "Transmission/Thrift/ServiceCommunicator.h"
#include "MixCastLog.h"
#include "Utility/CommandLineUtil.hpp"
#include "Tracking/TrackingUtil.hpp"
#include "Rendering/ExpCameraSpawner.h"
#include "Rendering/Projections/ExpVideoInputSpawner.h"
#include "Viewfinders/ExpViewfinderSpawner.h"

#include "Async/Async.h"
#include "Engine/Engine.h"
#include "HeadMountedDisplay.h"
#include "Engine/World.h"
#include "Camera/CameraComponent.h"
#include "GameFramework/PlayerController.h"
#include "Kismet/GameplayStatics.h"
#include "Runtime/Launch/Resources/Version.h"

#if ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION >= 18
#include "IXRTrackingSystem.h"
#else
#include "IHeadMountedDisplay.h"
#endif


FTransform AMixCastSdkActor::MixCastToWorld;

AMixCastSdkActor::AMixCastSdkActor( const FObjectInitializer& ObjectInitializer )
	: Super( ObjectInitializer )
{
	SpawnCollisionHandlingMethod = ESpawnActorCollisionHandlingMethod::AlwaysSpawn;

	PrimaryActorTick.bCanEverTick = true;
	PrimaryActorTick.TickGroup = TG_PostUpdateWork;
	PrimaryActorTick.bStartWithTickEnabled = true;
}

void AMixCastSdkActor::BeginPlay()
{
	Super::BeginPlay();

#if !UE_EDITOR
	if (SdkData::GetProjectSettings()->bRequireCommandLineFlag && !CommandLineUtil::IncludesMixCastFlag())
	{
		mod = nullptr;	//signal we didn't initialize
		Destroy();
		return;
	}
#else
	if (GetWorld()->IsNetMode(ENetMode::NM_DedicatedServer))
		return;
#endif

	mod = &IMixCast::Get();

	TrackingOrigin = NewObject<USceneComponent>(this);
	if (TrackingOrigin)
	{
		TrackingOrigin->Rename(TEXT("Tracking Origin"));
		TrackingOrigin->RegisterComponent();
		TrackingOrigin->AttachTo(GetRootComponent(), NAME_None);
	}

	CameraGroup = NewObject<UExpCameraSpawner>(this);
	if (CameraGroup)
	{
		CameraGroup->Rename(TEXT("Cameras"));
		CameraGroup->RegisterComponent();
		CameraGroup->AttachTo(TrackingOrigin, NAME_None);
	}

	VideoInputs = NewObject<UExpVideoInputSpawner>(this);
	if (VideoInputs)
	{
		VideoInputs->Rename(TEXT("VideoInputs"));
		VideoInputs->RegisterComponent();
		VideoInputs->AttachTo(TrackingOrigin, NAME_None);
	}

	Viewfinders = NewObject<UExpViewfinderSpawner>(this);
	if (Viewfinders)
	{
		Viewfinders->Rename(TEXT("Viewfinders"));
		Viewfinders->RegisterComponent();
		Viewfinders->AttachTo(TrackingOrigin, NAME_None);
	}

	ActiveChangedDelegate = mod->OnActiveChanged().AddUObject(this, &AMixCastSdkActor::HandleActiveChanged);
	CommandReceivedDelegate = mod->OnCommandReceived().AddUObject(this, &AMixCastSdkActor::HandleCommandReceived);

	((FMixCast*)mod)->RegisterSdkActor(this);
	HandleActiveChanged();
}
void AMixCastSdkActor::EndPlay(const EEndPlayReason::Type EndPlayReason)
{
	Super::EndPlay(EndPlayReason);

	if (mod == nullptr)
		return;

	mod->OnCommandReceived().Remove(CommandReceivedDelegate);
	mod->OnActiveChanged().Remove(ActiveChangedDelegate);

	CameraGroup->ClearSpawnedCameras();
	CameraGroup->DestroyComponent();
	CameraGroup = nullptr;

	VideoInputs->ClearSpawnedVideoInputs();
	VideoInputs->DestroyComponent();
	VideoInputs = nullptr;

	Viewfinders->ClearSpawnedViewfinders();
	Viewfinders->DestroyComponent();
	Viewfinders = nullptr;

	TrackingOrigin->DestroyComponent();
	TrackingOrigin = nullptr;

	((FMixCast*)mod)->UnregisterSdkActor(this);
}


void AMixCastSdkActor::HandleActiveChanged()
{
	if (!mod->IsActive())
	{
		VideoInputs->ClearSpawnedVideoInputs();
		CameraGroup->ClearSpawnedCameras();
	}
}

void AMixCastSdkActor::HandleCommandReceived(FString commandStr)
{
	OnCommandReceived(commandStr);
}
void AMixCastSdkActor::OnCommandReceived_Implementation(const FString& commandStr) { }

void AMixCastSdkActor::Tick(const float DeltaSeconds)
{
	if (mod == nullptr || !mod->IsActive())
		return;

	mod->OnBeforeFrame().Broadcast();

	TrackingOrigin->SetWorldTransform(MixCastToWorld);

	FServiceCommunicator* serviceCommunicator = ((FMixCast*)mod)->GetClientConnection();
	serviceCommunicator->SendSharedTextures();
	serviceCommunicator->UpdateCustomTrackedObjects();

	Viewfinders->SynchronizeSpawnedViewfinders();
	VideoInputs->SynchronizeSpawnedVideoInputs();
	CameraGroup->SynchronizeSpawnedCameras();

	CameraGroup->RenderCameras();
}
