/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#include "MixCast_SteamVR.h"

#include "MixCast.h"
#include "MixCastSdkActor.h"
#include "MixCast/Private/Tracking/TrackingUtil.hpp"
#include "MixCast_SteamVRLog.h"

#include "CoreMinimal.h"
#include "GameFramework/WorldSettings.h"
#include "Runtime/Launch/Resources/Version.h"
#include "HeadMountedDisplayFunctionLibrary.h"

#if ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION >= 19
#include "IXRTrackingSystem.h"
#endif

#if ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION >= 25
#include "SteamVR/Public/ISteamVRPlugin.h"
#elif ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION >= 20
#include "SteamVR/Public/ISteamVRPlugin.h"
#else
#include "Public/ISteamVRPlugin.h"
#endif



IMPLEMENT_MODULE(FMixCast_SteamVR, MixCast_SteamVR)

void FMixCast_SteamVR::StartupModule()
{
	trackingToPlayArea = FTransform::Identity;

	IMixCast& mod = IMixCast::Get();
	mixcastActiveDelegate = mod.OnActiveChanged().AddRaw(this, &FMixCast_SteamVR::HandleMixCastActiveChanged);
	trackingSpaceChangedDelegate = mod.OnTrackingSpaceChanged().AddRaw(this, &FMixCast_SteamVR::HandleTrackingSpaceChanged);
	beforeFrameDelegate = mod.OnBeforeFrame().AddRaw(this, &FMixCast_SteamVR::HandleMixCastAboutToRender);
}
void FMixCast_SteamVR::ShutdownModule()
{
	if (IMixCast::IsAvailable()) {
		IMixCast& mod = IMixCast::Get();
		mod.OnActiveChanged().Remove(mixcastActiveDelegate);
		mod.OnTrackingSpaceChanged().Remove(trackingSpaceChangedDelegate);
		mod.OnBeforeFrame().Remove(beforeFrameDelegate);
	}
}
void FMixCast_SteamVR::HandleMixCastActiveChanged()
{
	if (IMixCast::Get().IsActive())
		HandleTrackingSpaceChanged();
}
void FMixCast_SteamVR::HandleTrackingSpaceChanged()
{
	FTransform newTrans;
	if (CalculateStandingToPlayAreaTransform(newTrans))
		trackingToPlayArea = newTrans;
}
void FMixCast_SteamVR::HandleMixCastAboutToRender()
{
	if (!TrackingUtil::IsPlatformSteamVR())
		return;

	FTransform hmdToWorld, trackingOriginToHmd, trackingToStanding;
	bool success = CalculateHmdToWorldTransform(hmdToWorld);
	if (!success)
		return;

	success = CalculateTrackingOriginToHmdTransform(trackingOriginToHmd);
	if (!success)
		return;

	success = CalculateTrackingOriginToStandingTransform(trackingToStanding);
	if (!success)
		return;

	if (!firstTickComplete) {
		firstTickComplete = true;
		HandleTrackingSpaceChanged();
	}

	FTransform playAreaToWorld = trackingToPlayArea * trackingToStanding.Inverse() * trackingOriginToHmd.Inverse() * hmdToWorld;

	AMixCastSdkActor::UpdateMixCastToWorldTransform(playAreaToWorld);
}

bool FMixCast_SteamVR::CalculateHmdToWorldTransform(FTransform& trans)
{
#if ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION >= 19
	if (!GEngine->XRSystem.IsValid())
		return false;

	trans = GEngine->XRSystem->GetTrackingToWorldTransform();
	return true;
#else
	UWorld* curWorld = ((FMixCast*)&IMixCast::Get())->GetWorld();
	if (curWorld == NULL)
		return false;

	APlayerController* localPlayerController = GEngine->GetFirstLocalPlayerController(curWorld);
	if (localPlayerController == NULL)
		return false;

	APlayerCameraManager* playerCamManager = localPlayerController->PlayerCameraManager;
	if (playerCamManager == NULL)
		return false;

	trans = FTransform(playerCamManager->GetCameraRotation(), playerCamManager->GetCameraLocation());
	return true;
#endif
}

bool FMixCast_SteamVR::CalculateTrackingOriginToHmdTransform(FTransform& trans)
{
#if ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION >= 19
	trans = FTransform::Identity;	//can get the tracking origin on its own now
	return true;
#elif ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION >= 18
	FQuat trackedHmdRot;
	FVector trackedHmdPos;
	if (!GEngine->XRSystem.IsValid() || !GEngine->XRSystem->GetCurrentPose(IXRTrackingSystem::HMDDeviceId, trackedHmdRot, trackedHmdPos))
		return false;

	trans = FTransform(trackedHmdRot, trackedHmdPos);
	return true;
#else
	if (!GEngine->HMDDevice.IsValid() || !GEngine->HMDDevice->HasValidTrackingPosition())
		return false;

	FQuat trackedHmdRot;
	FVector trackedHmdPos;
	GEngine->HMDDevice->GetCurrentOrientationAndPosition(trackedHmdRot, trackedHmdPos);
	trans = FTransform(trackedHmdRot, trackedHmdPos);
	return true;
#endif
}

bool FMixCast_SteamVR::CalculateTrackingOriginToStandingTransform(FTransform& trans)
{
	if (UHeadMountedDisplayFunctionLibrary::GetTrackingOrigin() != EHMDTrackingOrigin::Eye) {
		trans = FTransform::Identity;
		return true;
	}

	UWorld* curWorld = ((FMixCast*)&IMixCast::Get())->GetWorld();
	vr::COpenVRContext& ctx = vr::OpenVRInternal_ModuleContext();
	if (curWorld == NULL || ctx.VRSystem() == NULL)
		return false;

	vr::HmdMatrix34_t steamSeatedToStanding = ctx.VRSystem()->GetSeatedZeroPoseToStandingAbsoluteTrackingPose();

	// Transform SteamVR Pose to Unreal Pose
	FMatrix Pose = ToFMatrix(steamSeatedToStanding);

	// Transform SteamVR Rotation Quaternion to a UE FRotator
	FQuat OrientationQuat;
	FQuat OrientationPose(Pose);
	OrientationQuat.X = -OrientationPose.Z;
	OrientationQuat.Y = OrientationPose.X;
	OrientationQuat.Z = OrientationPose.Y;
	OrientationQuat.W = -OrientationPose.W;

	FVector PositionPose = ((FVector(-Pose.M[3][2], Pose.M[3][0], Pose.M[3][1])) * curWorld->GetWorldSettings()->WorldToMeters);

	//OutOrientation = BaseOrientation.Inverse() * OutOrientation;
	OrientationQuat.Normalize();

	trans = FTransform(OrientationQuat, PositionPose);
	return true;
}

bool FMixCast_SteamVR::CalculateStandingToPlayAreaTransform(FTransform& trans)
{
	vr::COpenVRContext& ctx = vr::OpenVRInternal_ModuleContext();
	if (ctx.VRChaperone() == NULL)
		return false;

	uint32_t numBoundaryPoints = 0;
	ctx.VRChaperoneSetup()->GetLiveCollisionBoundsInfo(nullptr, &numBoundaryPoints);
	if (numBoundaryPoints == 0)
		return false;

	vr::HmdQuad_t* boundaryGeo = new vr::HmdQuad_t[numBoundaryPoints];
	if (!ctx.VRChaperoneSetup()->GetLiveCollisionBoundsInfo(boundaryGeo, &numBoundaryPoints) || numBoundaryPoints == 0)
	{
		delete[] boundaryGeo;
		return false;
	}

	UWorld* curWorld = ((FMixCast*)&IMixCast::Get())->GetWorld();
	float worldScale = curWorld->GetWorldSettings()->WorldToMeters;

	FVector averagePos;
	for (uint32_t i = 0; i < numBoundaryPoints; i++)
		averagePos += CONVERT_STEAMVECTOR_TO_FVECTOR(boundaryGeo[i].vCorners[0]);
	averagePos /= numBoundaryPoints;
	averagePos *= worldScale;

	FVector forwardPos = CONVERT_STEAMVECTOR_TO_FVECTOR(boundaryGeo[0].vCorners[0]) * worldScale;

	delete[] boundaryGeo;

	FQuat trackingCenterRot = (forwardPos - averagePos).ToOrientationQuat();
	trans = FTransform(trackingCenterRot, averagePos);
	return true;
}